import org.opencv.core.Core;
import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.Scalar;
import org.opencv.core.Size;

public class Pose {
	
	/**
	 * In the given camera coordinate system the axis conventions are different from the aircraft coordinate conventions.
	 * The left side represents the aircraft coordinate convention and the right site the corrosponding given camera coordinate axis
	 * x = z
	 * y = x
	 * z = y
	 * aircraft coordinate system
	 */
	
	private static Mat basisExchangeOpenCVToEuropeanCartesian;
	private static Mat basisExchangeEuropeanCartesianToOpenCV;
	static{
		
		/*
		 * Assumtation of the open-cv camera axis:
		 * y goes down.
		 * z goes forward
		 * x goes rightward
		 * 
		 * Assmutation of the world coordinate axis:
		 * x goes foreward
		 * z goes downward
		 * y goes leftward
		 * 
		 * x = z
		 * y = -x
		 * z = -y
		 * 
		 * */
		
		basisExchangeOpenCVToEuropeanCartesian = Mat.zeros(new Size(3,3),CvType.CV_64F);
		double[] row0 = new double[]{0,0,1};
		double[] row1 = new double[]{-1,0,0};
		double[] row2 = new double[]{0,1,0};
		basisExchangeOpenCVToEuropeanCartesian.put(0, 0, row0);
		basisExchangeOpenCVToEuropeanCartesian.put(1, 0, row1);
		basisExchangeOpenCVToEuropeanCartesian.put(2, 0, row2);
		
		basisExchangeEuropeanCartesianToOpenCV = basisExchangeOpenCVToEuropeanCartesian.inv();
	}
	
	private static double convertToDegree(double radians){
		return (radians * 180) / Math.PI;
	}
	
	private static double convertToRadians(double degrees){
		return (degrees * Math.PI) / 180.0;
	}
	
	private Mat rotation;
	private Mat translation;
	
	
	private double[] translationArr;
	private double[][] rotationArr;
	
	private Mat rotationC;
	private Mat translationC;
	private boolean isValid;
	
	public Pose(Mat rotation, Mat translation, boolean worldCoordinateInitialization){
		if(worldCoordinateInitialization){
			this.translation = translation;
			this.rotation = rotation;
			initRTArrays();
			/*compute the rotation matrix and tVec in the OpenCv coordinate system*/
			rotationC = new Mat();
			Core.gemm(basisExchangeEuropeanCartesianToOpenCV, rotation, 1, new Mat(), 0, rotationC);
			Core.gemm(rotationC, basisExchangeOpenCVToEuropeanCartesian, 1, new Mat(), 0, rotationC);
			
			translationC = new Mat();
			Core.gemm(basisExchangeEuropeanCartesianToOpenCV, translation, 1, new Mat(), 0, translationC);
			
		}else{
			initialize(rotation, translation);
		}
	}
	
	public Pose(Mat rotationC, Mat translationC) {
		super();
		initialize(rotationC, translationC);
	}
	
	private void initialize(Mat rotationC, Mat translationC){
		this.rotationC = rotationC;
		this.translationC = translationC;
		
		/*save the rotation matrix in world coordinates*/
		rotation = new Mat();
		/*the result of the rotation should be in world coordinates*/
		Core.gemm(basisExchangeOpenCVToEuropeanCartesian, rotationC, 1, new Mat(), 0, rotation);
		/*but the original input was in open-cv coordinates and so must be transformed*/
		Core.gemm(rotation, basisExchangeEuropeanCartesianToOpenCV, 1, new Mat(), 0, rotation);
		
		/*save the translation matrix in world coordinates*/
		translation = new Mat();
		Core.gemm(basisExchangeOpenCVToEuropeanCartesian, translationC, 1, new Mat(), 0, translation);
		initRTArrays();
	}
	
	private void initRTArrays(){
		/*save the values in an array*/
		rotationArr = new double[rotation.width()][];
		for(int i=0;i<rotation.width(); i++){
			rotationArr[i] = new double[rotation.height()];
			double[] value;
			for(int j=0; j<rotation.height();j++){
				value = rotation.get(i, j);
				rotationArr[i][j] = value[0];
			}
		}
		translationArr = new double[3];
		for(int i=0; i<translation.height(); i++){
			double[] value = translation.get(i, 0);
			translationArr[i] = value[0];
		}
	}
	
	/**
	 * http://nghiaho.com/?page_id=846
	 */
	
	/**
	 * @return the roll rotation around the x axis
	 */
	public double getRollRadians(){
		return Math.atan2(rotationArr[2][1], rotationArr[2][2]);
	}
	
	/**
	 * @return the pitch rotation around the y axis
	 */
	public double getPitchRadians(){
		return Math.atan2(-1*rotationArr[2][0], Math.sqrt( Math.pow(rotationArr[2][1], 2) + Math.pow(rotationArr[2][2], 2) ));
	}
	
	/**
	 * @return the pitch rotation around the z axis
	 */
	public double getYawRadians(){
		return Math.atan2(rotationArr[1][0], rotationArr[0][0]);
	}
	
	/**
	 * @return the roll rotation around the x axis
	 */
	public double getRoll(){
		return convertToDegree(getRollRadians());
	}
	
	/**
	 * @return the pitch rotation around the y axis
	 */
	public double getPitch(){
		return convertToDegree(getPitchRadians());
	}
	
	/**
	 * @return the pitch rotation around the z axis
	 */
	public double getYaw(){
		return convertToDegree(getYawRadians());
	}
	
	public Mat getRotation() {
		return rotation;
	}
	
	public Mat getTranslation() {
		return translation;
	}
	
	public double[] getTranslationArr() {
		return translationArr;
	}
	
	public double[][] getRotationArr() {
		return rotationArr;
	}
	
	public Mat getRotationOpenCv() {
		return rotationC;
	}
	
	public Mat getTranslationOpenCv() {
		return translationC;
	}
	
	
	public boolean isValid(){
		return isValid;
	}
	
	public void setValid(boolean isValid) {
		this.isValid = isValid;
	}
	
	public void getInverseProjectionOpenCv(Mat rMat, Mat tVec){
		computeInversePorjection(rotationC, translationC, rMat, tVec);
	}
	public void getInverseProjection(Mat rMat, Mat tVec){
		computeInversePorjection(rotation, translation, rMat, tVec);
	}
	
	private void computeInversePorjection(Mat rmatIn, Mat tvecIn, Mat rmatOut, Mat tvecOut){
		Mat rotation = rmatIn;
		Mat negRotation = new Mat();
		
		System.out.println("rotation-size:" + rotation.size());
		System.out.println("rotation:" + rotation.dump());
		rotation = rotation.t();
		Core.multiply(rotation, new Scalar(-1), negRotation);
		
		Core.gemm(negRotation, tvecIn, 1, new Mat(), 0, tvecOut);
		System.out.println("rotation-t:" + rotation.dump());
		System.out.println("neg-rotation:" + negRotation.dump());
		System.out.println("tvec-in:" + tvecIn.dump());
		System.out.println("tvec-out:" + tvecOut.dump());
		rotation.copyTo(rmatOut);
	}
	
	double getDx(){
		return translationArr[0];
	}
	double getDy(){
		return translationArr[1];
	}
	double getDz(){
		return translationArr[2];
	}
}