import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.opencv.core.CvType;
import org.opencv.core.Mat;
import org.opencv.core.MatOfDouble;
import org.opencv.core.Point;
import org.opencv.core.Point3;
import org.opencv.core.Size;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

public class XmlInterface {
	/*
	 * http://stackoverflow.com/questions/7373567/java-how-to-read-and-write-xml-files
	 * */
	private DocumentBuilderFactory dbFactory;

	public XmlInterface() {
		dbFactory = DocumentBuilderFactory.newInstance();
	}

	public Dataset loadDataset(File xmlFile) throws SAXException, IOException, ParserConfigurationException{
		Document dom;
		DocumentBuilder db = dbFactory.newDocumentBuilder();
		dom = db.parse(xmlFile);
		Element datasetNode = dom.getDocumentElement();
		
		
		Dataset dataset = new Dataset(new LinkedList<Picture>(), parseCameraMatrix(datasetNode), parseDistortionMatrix(datasetNode));

		NodeList pictureNodes = datasetNode.getElementsByTagName("picture");
		for(int i=0;i<pictureNodes.getLength();i++){
			Node pictureNode = pictureNodes.item(i);
			dataset.addPicture(loadPicture(pictureNode));
		}
		
		return dataset;
	}
	private Mat parseCameraMatrix(Element datasetNode) {
		String cameraMatrixString = datasetNode.getAttribute("camera-matrix");
		StringTokenizer tokenizer = new StringTokenizer(cameraMatrixString, ",", false);
		double[] values = new double[9];
		int i=0;
		while(tokenizer.hasMoreTokens()){
			values[i++] = Double.parseDouble(tokenizer.nextToken().trim());
		}
		Mat mat = new Mat(new Size(3, 3), CvType.CV_64F);
		mat.put(0,0,values);
		mat.reshape(0, 3);
		return mat;
	}

	private MatOfDouble parseDistortionMatrix(Element datasetNode) {
		String cameraMatrixString = datasetNode.getAttribute("distortion-matrix");
		StringTokenizer tokenizer = new StringTokenizer(cameraMatrixString, ",", false);
		double[] values = new double[5];
		int i=0;
		while(tokenizer.hasMoreTokens()){
			values[i++] = Double.parseDouble(tokenizer.nextToken().trim());
		}
		MatOfDouble mat = new MatOfDouble();
		mat.fromArray(values);
		return mat;
	}

	private Picture loadPicture(Node pictureNode) {
		NamedNodeMap attributes = pictureNode.getAttributes();
		String filename = attributes.getNamedItem("filename").getTextContent();
		
		LinkedList<LED> leds = new LinkedList<LED>();
		
		NodeList childNodes = pictureNode.getChildNodes();
		double expectedDx = 0;
		double expectedDy = 0;
		double expectedDz = 0;
		double  expectedRoll = 0;
		double expectedPitch = 0;
		double expectedYaw = 0;
		for(int i=0;i<childNodes.getLength(); i++){
			Node pictureChildNode = childNodes.item(i);
			if(pictureChildNode.getNodeName().equals("led")){
				int ledNumber = Integer.parseInt(pictureChildNode.getAttributes().getNamedItem("led-id").getTextContent());
				NodeList ledChildNodes = pictureChildNode.getChildNodes();
				Point imageCoordinate = null;
				Point3 worldCoordinate = null;
				for(int j=0;j<ledChildNodes.getLength();j++){
					Node ledChildNode = ledChildNodes.item(j);
					double x,y,z;
					switch (ledChildNode.getNodeName()) {
					case "image-coordinates":
						x = Double.parseDouble(ledChildNode.getAttributes().getNamedItem("x").getTextContent());
						y = Double.parseDouble(ledChildNode.getAttributes().getNamedItem("y").getTextContent());
						imageCoordinate = new Point(x, y);
						break;
					case "world-coordinates":
						x = Double.parseDouble(ledChildNode.getAttributes().getNamedItem("x").getTextContent());
						y = Double.parseDouble(ledChildNode.getAttributes().getNamedItem("y").getTextContent());
						z = Double.parseDouble(ledChildNode.getAttributes().getNamedItem("z").getTextContent());
						worldCoordinate = new Point3(x,y,z);
						
						break;

					default:
						break;
					}
				}
				leds.add(new LED(ledNumber, imageCoordinate, worldCoordinate));
			}
			if(pictureChildNode.getNodeName().equals("expectation")){
				expectedDx = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("dx").getTextContent());
				expectedDy = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("dy").getTextContent());
				expectedDz = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("dz").getTextContent());

				expectedRoll = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("roll").getTextContent());
				expectedPitch = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("pitch").getTextContent());
				expectedYaw = Double.parseDouble(pictureChildNode.getAttributes().getNamedItem("yaw").getTextContent());
			}
		}
		
		return new Picture(filename, leds, expectedDx, expectedDy, expectedDz, expectedRoll, expectedPitch, expectedYaw);
	}

	public void storeDataset(Dataset dataset, File xmlFile) throws ParserConfigurationException, TransformerFactoryConfigurationError, FileNotFoundException, TransformerException{

		DocumentBuilder db = dbFactory.newDocumentBuilder();
		Document dom = db.newDocument();
		Element datasetNode = dom.createElement("dataset");
		
		datasetNode.setAttribute("distortion-matrix", matToString(dataset.getDistortionMatrix()));
		datasetNode.setAttribute("camera-matrix", matToString(dataset.getCameraMatrix()));
		
		for(Picture picture:dataset.getPictures()){
			Element pictureNode = createPictureNode(picture, dom);
			datasetNode.appendChild(pictureNode);
		}

		dom.appendChild(datasetNode);

		Transformer tr = TransformerFactory.newInstance().newTransformer();
		tr.setOutputProperty(OutputKeys.INDENT, "yes");
		tr.setOutputProperty(OutputKeys.METHOD, "xml");
		tr.setOutputProperty(OutputKeys.ENCODING, "UTF-8");

		tr.transform(new DOMSource(dom), 
				new StreamResult(new FileOutputStream(xmlFile)));
	}

	private String matToString(Mat mat) {
		StringBuilder sb = new StringBuilder();
		for(int row=0;row<mat.rows(); row++){
			for(int col=0; col<mat.cols(); col++){
				double[] val = mat.get(row, col);
				sb.append(val[0]);
				sb.append(", ");
			}
		}
		return sb.substring(0,sb.length()-2).toString();
	}

	private Element createPictureNode(Picture picture, Document dom){
		Element pictureNode = dom.createElement("picture");
		/*configure attributes*/
		pictureNode.setAttribute("filename", picture.getFilename());
		pictureNode.appendChild(createExpectationNode(picture, dom));
		for(LED led:picture.getLeds()){
			Element ledNode = createLedNode(led, dom);
			pictureNode.appendChild(ledNode);
		}
		return pictureNode;
	}

	private Element createLedNode(LED led, Document dom){
		Element ledNode = dom.createElement("led");
		ledNode.setAttribute("led-id", String.valueOf(led.getNumber()));
		ledNode.appendChild(createImageCoordinatesNode(led.getImageCoordinate(), dom));
		ledNode.appendChild(createWorldCoordinatesNode(led.getWorldCoordinate(), dom));
		return ledNode;

	}
	private Element createExpectationNode(Picture picture, Document dom){
		Element expectationNode = dom.createElement("expectation");
		/*configure attributes*/
		expectationNode.setAttribute("dx", String.valueOf(picture.getExpectedDx()));
		expectationNode.setAttribute("dy", String.valueOf(picture.getExpectedDy()));
		expectationNode.setAttribute("dz", String.valueOf(picture.getExpectedDz()));
		expectationNode.setAttribute("roll", String.valueOf(picture.getExpectedRoll()));
		expectationNode.setAttribute("pitch", String.valueOf(picture.getExpectedPitch()));
		expectationNode.setAttribute("yaw", String.valueOf(picture.getExpectedYaw()));
		return expectationNode;
	}

	private Element createWorldCoordinatesNode(Point3 coords, Document dom){
		Element wordCoordinatesNode = dom.createElement("world-coordinates");
		/*configure attributes*/
		wordCoordinatesNode.setAttribute("x", String.valueOf(coords.x));
		wordCoordinatesNode.setAttribute("y", String.valueOf(coords.y));
		wordCoordinatesNode.setAttribute("z", String.valueOf(coords.z));
		return wordCoordinatesNode;
	}
	private Element createImageCoordinatesNode(Point coords, Document dom){
		Element imageCoordinatesNode = dom.createElement("image-coordinates");
		/*configure attributes*/
		imageCoordinatesNode.setAttribute("x", String.valueOf(coords.x));
		imageCoordinatesNode.setAttribute("y", String.valueOf(coords.y));
		return imageCoordinatesNode;
	}
}
