package de.tu_darmstadt.fsr.headtracker;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import org.opencv.core.Mat;
import org.opencv.core.MatOfFloat;
import org.opencv.core.Point3;

public class TrainingData implements Serializable{
	
	private static final long serialVersionUID = 1L;
	
	/**
	 * An array of trained histograms.
	 * The index of the array represents the corresponding LED id.
	 * The sub array has the length 3 and contains 3 histograms.
	 * One for the each r,g,b channel.
	 * */
	private transient Mat[][] hists;
	private float[][][] histsSerializable;
	

	/**
	 * An array containing the minimum possible distance in y direction of an LED neighbor
	 * for the LED with the given index
	 * Neighbors must always be on the right side of an LED.
	 * So the last LED does not have a neighbor.
	 * */
	private int[] minDistY;
	
	/**
	 * An array containing the minimum possible distance in x direction of an LED neighbor
	 * for the LED with the given index
	 * Neighbors must always be on the right side of an LED.
	 * So the last LED does not have a neighbor.
	 * */
	private int[] minDistX;
	
	/**
	 * An array containing the maximum possible distance in x direction of an LED neighbor
	 * for the LED with the given index
	 * Neighbors must always be on the right side of an LED.
	 * So the last LED does not have a neighbor.
	 * */
	private int[] maxDistY;
	
	/**
	 * An array containing the maximum possible distance in y direction of an LED neighbor
	 * for the LED with the given index
	 * Neighbors must always be on the right side of an LED.
	 * So the last LED does not have a neighbor.
	 * */
	private int[] maxDistX;

	/**
	 * An array containing the 3D world coordinates of every LED
	 */
	private transient Point3[] worldCoordinates;
	private double[][] worldCoordinatesSerializable;
	
	public TrainingData(Mat[][] hists, int[] minDistX, int[] minDistY,
			int[] maxDistX, int[] maxDistY, Point3[] worldCoordinates) {
		super();
		this.hists = hists;
		this.minDistY = minDistY;
		this.minDistX = minDistX;
		this.maxDistY = maxDistY;
		this.maxDistX = maxDistX;
		this.worldCoordinates = worldCoordinates;
	}

	public Mat[][] getHists() {
		return hists;
	}
	
	public int[] getMinDistX() {
		return minDistX;
	}
	
	public int[] getMinDistY() {
		return minDistY;
	}
	
	public int[] getMaxDistX() {
		return maxDistX;
	}
	
	public int[] getMaxDistY() {
		return maxDistY;
	}
	
	private void readObject(ObjectInputStream input)
			throws IOException, ClassNotFoundException {
		// deserialize the non-transient data members first;
		input.defaultReadObject();
		
		hists = new Mat[histsSerializable.length][];
		for(int i=0;i<histsSerializable.length;i++){
			hists[i] = new Mat[histsSerializable[i].length];
			for(int j=0;j<histsSerializable[i].length;j++){
				hists[i][j] = new MatOfFloat(histsSerializable[i][j]);
			}
		}
		
		worldCoordinates = new Point3[worldCoordinatesSerializable.length];
		for(int i=0; i<worldCoordinatesSerializable.length; i++){
			Point3 point3 = new Point3(
					worldCoordinatesSerializable[i][0],
					worldCoordinatesSerializable[i][1],
					worldCoordinatesSerializable[i][2]
			);
			worldCoordinates[i] = point3;
		}
	}
	private void writeObject(ObjectOutputStream output)
			throws IOException, ClassNotFoundException {
		// serialize the non-transient data members first;
		histsSerializable = new float[hists.length][][];
		for(int i=0;i<hists.length;i++){
			histsSerializable[i]=new float[hists[i].length][];
			for(int j=0;j<hists[i].length;j++){
				Mat mat = hists[i][j];
				histsSerializable[i][j] = new float[mat.height()];
				mat.get(0, 0, histsSerializable[i][j]);
			}
		}
		
		worldCoordinatesSerializable = new double[worldCoordinates.length][];
		for(int i=0; i<worldCoordinatesSerializable.length; i++){
			Point3 point3 = worldCoordinates[i];
			worldCoordinatesSerializable[i] = new double[3];
			worldCoordinatesSerializable[i][0] = point3.x;
			worldCoordinatesSerializable[i][1] = point3.y;
			worldCoordinatesSerializable[i][2] = point3.z;
		}
		
	    output.defaultWriteObject();
	}

	public int getLEDsCount() {
		return hists.length;
	}

	public Point3 getWorldCoordinate(int led) {
		return worldCoordinates[led];
	}
	
	public Point3[] getWorldCoordinates() {
		return worldCoordinates;
	}
}